From 015b82d9a8987f613c15671e480fa7b3c92bd5f5 Mon Sep 17 00:00:00 2001
From: "Dmitriy A. Gerasimov" <dmitriy.gerasimov@demlabs.net>
Date: Tue, 16 Feb 2021 00:59:33 +0700
Subject: [PATCH] [+] In-memory cache for dap_http objects [*] Some refactoring
 here, mostly renames and tabs

---
 dap-sdk/net/server/http_server/dap_http.c     | 133 +++++----
 .../net/server/http_server/dap_http_cache.c   |  69 +++++
 .../net/server/http_server/dap_http_simple.c  |  13 +
 .../http_server/http_client/dap_http_client.c | 267 ++++++++++--------
 .../http_server/http_client/dap_http_header.c |  41 ++-
 .../http_client/include/dap_http_client.h     |  17 +-
 .../http_client/include/dap_http_header.h     |  47 +--
 .../net/server/http_server/include/dap_http.h |   5 +
 .../http_server/include/dap_http_cache.h      |  40 +++
 .../http_server/include/dap_http_simple.h     |   2 +
 10 files changed, 411 insertions(+), 223 deletions(-)
 create mode 100644 dap-sdk/net/server/http_server/dap_http_cache.c
 create mode 100644 dap-sdk/net/server/http_server/include/dap_http_cache.h

diff --git a/dap-sdk/net/server/http_server/dap_http_cache.c b/dap-sdk/net/server/http_server/dap_http_cache.c
new file mode 100644
index 0000000000..5980b5e030
--- /dev/null
+++ b/dap-sdk/net/server/http_server/dap_http_cache.c
@@ -0,0 +1,69 @@
+/*
+ * Authors:
+ * Dmitriy A. Gearasimov <gerasimov.dmitriy@demlabs.net>
+ * DeM Labs Ltd.   https://demlabs.net
+ * Copyright  (c) 2021
+ * All rights reserved.
+
+ This file is part of DAP SDK the open source project
+
+    DAP SDK is free software: you can redistribute it and/or modify
+    it under the terms of the GNU General Public License as published by
+    the Free Software Foundation, either version 3 of the License, or
+    (at your option) any later version.
+
+    DAP SDK is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+    GNU General Public License for more details.
+
+    You should have received a copy of the GNU General Public License
+    along with any DAP SDK based project.  If not, see <http://www.gnu.org/licenses/>.
+*/
+#include "utlist.h"
+#include "dap_http.h"
+#include "dap_http_cache.h"
+
+#define LOG_TAG "http_cache"
+
+dap_http_cache_t * dap_http_cache_update(struct dap_http_url_proc * a_url_proc, byte_t * a_body, size_t a_body_size,
+                                         dap_http_header_t * a_headers, time_t a_ts_expire )
+{
+    dap_http_cache_t * l_ret = DAP_NEW_Z(dap_http_cache_t);
+    if(a_body_size){
+        l_ret->body = DAP_NEW_SIZE(byte_t,a_body_size);
+        memcpy(l_ret->body,a_body,a_body_size);
+        l_ret->body_size = a_body_size;
+    }
+    l_ret->headers =  dap_http_headers_dup( a_headers);
+    l_ret->ts_expire = a_ts_expire;
+    l_ret->url_proc = a_url_proc;
+    pthread_rwlock_wrlock(&a_url_proc->cache_rwlock);
+    dap_http_cache_delete(a_url_proc->cache);
+    a_url_proc->cache = l_ret;
+    pthread_rwlock_unlock(&a_url_proc->cache_rwlock);
+    return l_ret;
+}
+
+/**
+ * @brief dap_http_cache_delete
+ * @param a_http_cache
+ */
+void dap_http_cache_delete(dap_http_cache_t * a_http_cache)
+{
+   if (a_http_cache){
+       if(a_http_cache->body)
+           DAP_DELETE(a_http_cache->body);
+       dap_http_header_t *l_hdr=NULL, *l_tmp=NULL;
+
+       DL_FOREACH_SAFE(a_http_cache->headers,l_hdr,l_tmp){
+           DL_DELETE(a_http_cache->headers,l_hdr);
+           if(l_hdr->name)
+               DAP_DELETE(l_hdr->name);
+           if(l_hdr->value)
+               DAP_DELETE(l_hdr->value);
+           DAP_DELETE(l_hdr);
+       }
+       DAP_DELETE(a_http_cache);
+   }
+}
diff --git a/dap-sdk/net/server/http_server/dap_http_simple.c b/dap-sdk/net/server/http_server/dap_http_simple.c
index a159c2c4f5..1daea3f597 100644
--- a/dap-sdk/net/server/http_server/dap_http_simple.c
+++ b/dap-sdk/net/server/http_server/dap_http_simple.c
@@ -439,6 +439,19 @@ size_t dap_http_simple_reply(dap_http_simple_t *a_http_simple, void *a_data, siz
     return l_data_copy_size;
 }
 
+/**
+ * @brief dap_http_simple_make_cache_from_reply
+ * @param a_http_simple
+ * @param a_ts_expire
+ */
+dap_http_cache_t * dap_http_simple_make_cache_from_reply(dap_http_simple_t * a_http_simple, time_t a_ts_expire  )
+{
+    return dap_http_cache_update(a_http_simple->http_client->http->url_proc,
+                                 a_http_simple->reply_byte,
+                                 a_http_simple->reply_size,
+                                 a_http_simple->http_client->out_headers, a_ts_expire);
+}
+
 /**
  * @brief dap_http_simple_reply_f
  * @param shs
diff --git a/dap-sdk/net/server/http_server/http_client/dap_http_client.c b/dap-sdk/net/server/http_server/http_client/dap_http_client.c
index 64f1b00468..a3ae2a556c 100644
--- a/dap-sdk/net/server/http_server/http_client/dap_http_client.c
+++ b/dap-sdk/net/server/http_server/http_client/dap_http_client.c
@@ -46,6 +46,9 @@
 
 #define LOG_TAG "dap_http_client"
 
+static bool s_request_line_parse( dap_http_client_t *cl_ht, char *buf, size_t buf_length );
+
+
 /**
  * @brief dap_http_client_init Init HTTP client module
  * @return  Zero if ok others if not
@@ -66,50 +69,62 @@ void dap_http_client_deinit( )
 
 /**
  * @brief dap_http_client_new Creates HTTP client's internal structure
- * @param cl HTTP Client instance
- * @param arg Additional argument (usualy not used)
+ * @param a_esocket ESocket instance
+ * @param a_arg Additional argument (usualy not used)
  */
-void dap_http_client_new( dap_events_socket_t *cl, void *arg )
+void dap_http_client_new( dap_events_socket_t *a_esocket, void *a_arg )
 {
-    (void) arg;
-
+    (void) a_arg;
 
-    cl->_inheritor = DAP_NEW_Z( dap_http_client_t );
 
-    dap_http_client_t *cl_ht = DAP_HTTP_CLIENT( cl );
-    cl_ht->esocket = cl;
-    cl_ht->http = DAP_HTTP( cl->server );
-    cl_ht->state_read = DAP_HTTP_CLIENT_STATE_START;
-    cl_ht->state_write = DAP_HTTP_CLIENT_STATE_NONE;
+    a_esocket->_inheritor = DAP_NEW_Z( dap_http_client_t );
 
+    dap_http_client_t *l_http_client = DAP_HTTP_CLIENT( a_esocket );
+    l_http_client->esocket = a_esocket;
+    l_http_client->http = DAP_HTTP( a_esocket->server );
+    l_http_client->state_read = DAP_HTTP_CLIENT_STATE_START;
+    l_http_client->state_write = DAP_HTTP_CLIENT_STATE_NONE;
+
+    pthread_rwlock_rdlock(&l_http_client->http->url_proc->cache_rwlock);
+    if(l_http_client->http->url_proc->cache){
+        if (l_http_client->http->url_proc->cache->ts_expire < time(NULL) )
+            l_http_client->out_headers = dap_http_headers_dup(l_http_client->http->url_proc->cache->headers);
+        else{
+            pthread_rwlock_unlock(&l_http_client->http->url_proc->cache_rwlock);
+            pthread_rwlock_wrlock(&l_http_client->http->url_proc->cache_rwlock);
+            dap_http_cache_delete(l_http_client->http->url_proc->cache);
+            l_http_client->http->url_proc->cache = NULL;
+        }
+    }
+    pthread_rwlock_unlock(&l_http_client->http->url_proc->cache_rwlock);
     return;
 }
 
 /**
  * @brief dap_http_client_delete
- * @param cl HTTP Client instance
- * @param arg Additional argument (usualy not used)
+ * @param a_esocket HTTP Client instance's esocket
+ * @param a_arg Additional argument (usualy not used)
  */
-void dap_http_client_delete( dap_events_socket_t * cl, void *arg )
+void dap_http_client_delete( dap_events_socket_t * a_esocket, void *a_arg )
 {
-    dap_http_client_t *cl_ht = DAP_HTTP_CLIENT( cl );
-    if (cl_ht == NULL){ // Client is in proc callback in another thread so we don't delete it
+    (void) a_arg;
+    dap_http_client_t *l_http_client = DAP_HTTP_CLIENT( a_esocket );
+    if (l_http_client == NULL){ // Client is in proc callback in another thread so we don't delete it
         return;
     }
-    while( cl_ht->in_headers )
-        dap_http_header_remove( &cl_ht->in_headers, cl_ht->in_headers );
+    while( l_http_client->in_headers )
+        dap_http_header_remove( &l_http_client->in_headers, l_http_client->in_headers );
 
-    while( cl_ht->out_headers )
-        dap_http_header_remove( &cl_ht->out_headers, cl_ht->out_headers );
+    while( l_http_client->out_headers )
+        dap_http_header_remove( &l_http_client->out_headers, l_http_client->out_headers );
 
-    if( cl_ht->proc ) {
-        if( cl_ht->proc->delete_callback ) {
-          cl_ht->proc->delete_callback( cl_ht, NULL );
+    if( l_http_client->proc ) {
+        if( l_http_client->proc->delete_callback ) {
+          l_http_client->proc->delete_callback( l_http_client, NULL );
         }
     }
-    DAP_DEL_Z(cl_ht->_inheritor)
+    DAP_DEL_Z(l_http_client->_inheritor)
 
-  (void) arg;
 }
 
 
@@ -121,12 +136,12 @@ void dap_http_client_delete( dap_events_socket_t * cl, void *arg )
  */
 
 #if 1
-int detect_end_of_line( const char *buf, size_t max_size )
+static int detect_end_of_line( const char *a_buf, size_t a_max_size )
 {
   size_t i;
 
-  for( i = 0; i < max_size; i++ ) {
-    if ( buf[i] == '\n' ) {
+  for( i = 0; i < a_max_size; i++ ) {
+    if ( a_buf[i] == '\n' ) {
       return i;
     }
   }
@@ -135,7 +150,7 @@ int detect_end_of_line( const char *buf, size_t max_size )
 }
 #endif
 
-char  *z_basename( char *path, uint32_t len )
+static char  *z_basename( char *path, uint32_t len )
 {
   if ( !len )
     len = strlen( path );
@@ -156,7 +171,7 @@ char  *z_basename( char *path, uint32_t len )
   return ptr;
 }
 
-int32_t  z_dirname( char *path, uint32_t len )
+static int32_t  z_dirname( char *path, uint32_t len )
 {
   if ( !len )
     len = strlen( path );
@@ -180,7 +195,7 @@ int32_t  z_dirname( char *path, uint32_t len )
   return len;
 }
 
-int32_t  z_rootdirname( char *path, uint32_t len )
+static int32_t  z_rootdirname( char *path, uint32_t len )
 {
   if ( !len )
     len = strlen( path );
@@ -207,55 +222,55 @@ int32_t  z_rootdirname( char *path, uint32_t len )
 }
 
 /**
- * @brief dap_http_request_line_parse
- * @param cl_ht
- * @param buf
- * @param buf_length
+ * @brief s_request_line_parse
+ * @param a_http_client
+ * @param a_buf
+ * @param a_buf_length
  * @return
  */
-bool dap_http_request_line_parse( dap_http_client_t *cl_ht, char *buf, size_t buf_length )
+static bool s_request_line_parse( dap_http_client_t *a_http_client, char *a_buf, size_t a_buf_length )
 {
-  size_t pos;
-  size_t pos_kw_begin = 0;
+  size_t l_pos;
+  size_t l_pos_kw_begin = 0;
 
-  enum parse_state { PS_START = 0, PS_ACTION = 1, PS_URL = 2, PS_TYPE = 3, PS_VER_MAJOR = 4, PS_VER_MINOR = 5 }  p_st = PS_ACTION;
+  enum parse_state { PS_START = 0, PS_ACTION = 1, PS_URL = 2, PS_TYPE = 3, PS_VER_MAJOR = 4, PS_VER_MINOR = 5 }  l_parse_state = PS_ACTION;
 
   log_it( L_NOTICE, "dap_http_request_line_parse" );
 
-  for( pos = 0; pos < buf_length; pos ++ ) {
+  for( l_pos = 0; l_pos < a_buf_length; l_pos ++ ) {
 
-    if ( buf[pos] == '\n' )
+    if ( a_buf[l_pos] == '\n' )
       break;
 
-    if ( buf[pos] == ' ' || buf[pos] == '\t' ) {
+    if ( a_buf[l_pos] == ' ' || a_buf[l_pos] == '\t' ) {
 
-      switch( p_st ) {
+      switch( l_parse_state ) {
       case PS_ACTION:
       {
-        size_t c_size = pos - pos_kw_begin;
-        if ( c_size + 1 > sizeof(cl_ht->action) )
-          c_size = sizeof( cl_ht->action ) - 1;
+        size_t c_size = l_pos - l_pos_kw_begin;
+        if ( c_size + 1 > sizeof(a_http_client->action) )
+          c_size = sizeof( a_http_client->action ) - 1;
 
-        memcpy( cl_ht->action, buf + pos_kw_begin, c_size );
-        cl_ht->action[c_size] = 0;
-        log_it( L_WARNING, "Input: action '%s' pos=%u pos_kw_begin=%u", cl_ht->action, (uint32_t)pos, (uint32_t)pos_kw_begin );
+        memcpy( a_http_client->action, a_buf + l_pos_kw_begin, c_size );
+        a_http_client->action[c_size] = 0;
+        log_it( L_WARNING, "Input: action '%s' pos=%u pos_kw_begin=%u", a_http_client->action, (uint32_t)l_pos, (uint32_t)l_pos_kw_begin );
 
-        p_st = PS_URL;
-        pos_kw_begin = pos + 1;
+        l_parse_state = PS_URL;
+        l_pos_kw_begin = l_pos + 1;
       }
       break;
 
       case PS_URL:
       {
-        size_t c_size = pos - pos_kw_begin;
-        if ( c_size + 1 > sizeof(cl_ht->action) )
-          c_size = sizeof( cl_ht->url_path ) - 1;
-
-        memcpy( cl_ht->url_path, buf + pos_kw_begin, c_size );
-        cl_ht->url_path[c_size] = 0;
-        log_it( L_WARNING, "Input: url '%s' pos=%lu pos_kw_begin=%lu", cl_ht->url_path, (uint32_t)pos, (uint32_t)pos_kw_begin );
-        p_st = PS_TYPE;
-        pos_kw_begin = pos + 1;
+        size_t c_size = l_pos - l_pos_kw_begin;
+        if ( c_size + 1 > sizeof(a_http_client->action) )
+          c_size = sizeof( a_http_client->url_path ) - 1;
+
+        memcpy( a_http_client->url_path, a_buf + l_pos_kw_begin, c_size );
+        a_http_client->url_path[c_size] = 0;
+        log_it( L_WARNING, "Input: url '%s' pos=%lu pos_kw_begin=%lu", a_http_client->url_path, (uint32_t)l_pos, (uint32_t)l_pos_kw_begin );
+        l_parse_state = PS_TYPE;
+        l_pos_kw_begin = l_pos + 1;
         break;
       }
       break;
@@ -266,52 +281,52 @@ bool dap_http_request_line_parse( dap_http_client_t *cl_ht, char *buf, size_t bu
     }
   } // for
 
-  if ( pos_kw_begin < buf_length && p_st == PS_TYPE ) {
+  if ( l_pos_kw_begin < a_buf_length && l_parse_state == PS_TYPE ) {
 
-    size_t c_size;
+    size_t l_c_size;
 
-    char *end = memchr( buf + pos_kw_begin, '/', buf_length - pos_kw_begin );
+    char *end = memchr( a_buf + l_pos_kw_begin, '/', a_buf_length - l_pos_kw_begin );
 
-    if ( end && end < buf + buf_length ) {
+    if ( end && end < a_buf + a_buf_length ) {
 
-      c_size = end - (buf + pos_kw_begin);
+      l_c_size = end - (a_buf + l_pos_kw_begin);
       //TODO get version here
       //end = memchr( buf + pos_kw_begin, '/', buf_length - pos_kw_begin );
 
     }
     else
-      c_size = buf_length - pos_kw_begin;
+      l_c_size = a_buf_length - l_pos_kw_begin;
 
-    if ( c_size + 1 > sizeof(cl_ht->in_content_type) )
-       c_size = sizeof(cl_ht->in_content_type) - 1;
+    if ( l_c_size + 1 > sizeof(a_http_client->in_content_type) )
+       l_c_size = sizeof(a_http_client->in_content_type) - 1;
 
-    memcpy( cl_ht->in_content_type, buf + pos_kw_begin, c_size );
-    cl_ht->in_content_type[c_size] = 0;
+    memcpy( a_http_client->in_content_type, a_buf + l_pos_kw_begin, l_c_size );
+    a_http_client->in_content_type[l_c_size] = 0;
 
-    log_it( L_WARNING, "Input: type '%s' pos=%lu pos_kw_begin=%lu", cl_ht->in_content_type, (uint32_t)pos, (uint32_t)pos_kw_begin );
+    log_it( L_WARNING, "Input: type '%s' pos=%lu pos_kw_begin=%lu", a_http_client->in_content_type, (uint32_t)l_pos, (uint32_t)l_pos_kw_begin );
   }
 
-  return cl_ht->url_path[0] && cl_ht->action[0];
+  return a_http_client->url_path[0] && a_http_client->action[0];
 }
 
 /**
  * @brief s_report_error_and_restart
- * @param cl
- * @param cl_ht
+ * @param a_esocket
+ * @param a_http_client
  */
-static inline void s_report_error_and_restart( dap_events_socket_t *cl, dap_http_client_t *cl_ht )
+static inline void s_report_error_and_restart( dap_events_socket_t *a_esocket, dap_http_client_t *a_http_client )
 {
-  cl->buf_in_size = 0;
-  cl_ht->state_read = DAP_HTTP_CLIENT_STATE_NONE;
+    a_esocket->buf_in_size = 0;
+    a_http_client->state_read = DAP_HTTP_CLIENT_STATE_NONE;
 
-  dap_events_socket_set_readable_unsafe( cl_ht->esocket, false );
-  dap_events_socket_set_writable_unsafe( cl_ht->esocket, true );
+    dap_events_socket_set_readable_unsafe( a_http_client->esocket, false );
+    dap_events_socket_set_writable_unsafe( a_http_client->esocket, true );
 
-  cl_ht->reply_status_code = 505;
-  strcpy( cl_ht->reply_reason_phrase, "Error" );
-  cl_ht->state_write = DAP_HTTP_CLIENT_STATE_START;
+    a_http_client->reply_status_code = 505;
+    strcpy( a_http_client->reply_reason_phrase, "Error" );
+    a_http_client->state_write = DAP_HTTP_CLIENT_STATE_START;
 
-  return;
+    return;
 }
 
 /**
@@ -319,7 +334,7 @@ static inline void s_report_error_and_restart( dap_events_socket_t *cl, dap_http
  * @param cl HTTP Client instance
  * @param arg Additional argument (usualy not used)
  */
-void dap_http_client_read( dap_events_socket_t *a_esocket, void *arg )
+void dap_http_client_read( dap_events_socket_t *a_esocket, void *a_arg )
 {
     dap_http_client_t *l_http_client = DAP_HTTP_CLIENT( a_esocket );
 
@@ -358,7 +373,7 @@ void dap_http_client_read( dap_events_socket_t *a_esocket, void *arg )
                 l_buf_line[ eol + 2 ] = 0; // null terminate
 
                 // parse http_request_line
-                if ( !dap_http_request_line_parse(l_http_client, l_buf_line, eol + 1) ) {
+                if ( !s_request_line_parse(l_http_client, l_buf_line, eol + 1) ) {
                     log_it( L_WARNING, "Input: Wrong request line '%s'", l_buf_line );
                     s_report_error_and_restart( a_esocket, l_http_client );
                     break;
@@ -474,30 +489,41 @@ void dap_http_client_read( dap_events_socket_t *a_esocket, void *arg )
 
 /**
  * @brief dap_http_client_write Process write event
- * @param cl HTTP Client instance
- * @param arg Additional argument (usualy not used)
+ * @param a_esocket HTTP Client instance's esocket
+ * @param a_arg Additional argument (usualy not used)
  */
-void dap_http_client_write( dap_events_socket_t * cl, void *arg )
+void dap_http_client_write( dap_events_socket_t * a_esocket, void *a_arg )
 {
     //  log_it( L_DEBUG, "dap_http_client_write..." );
 
-    (void) arg;
-    dap_http_client_t *l_http_client = DAP_HTTP_CLIENT( cl );
+    (void) a_arg;
+    dap_http_client_t *l_http_client = DAP_HTTP_CLIENT( a_esocket );
     //log_it(L_WARNING,"HTTP client write callback in state %d",l_http_client->state_write);
 
     switch( l_http_client->state_write ) {
         case DAP_HTTP_CLIENT_STATE_NONE:
             return;
         case DAP_HTTP_CLIENT_STATE_START:{
-            if ( l_http_client->proc )
-                if ( l_http_client->proc->headers_write_callback )
-                    l_http_client->proc->headers_write_callback( l_http_client, NULL );
+            if ( l_http_client->proc ){
+                // We check out_headers because if they are - we send only cached headers and don't call headers_write_callback at all
+                if ( l_http_client->out_headers==NULL && l_http_client->proc->headers_write_callback ){
+                        l_http_client->proc->headers_write_callback( l_http_client, NULL );
+                        dap_http_client_out_header_generate( l_http_client );
+                }else if (l_http_client->out_headers){
+                    l_http_client->reply_status_code = Http_Status_OK; // Cached data are always OK... for now.
+                    //TODO: make cached reply status code
+                }
+            }
+            char buf[1024];
+            time_t current_time = time( NULL );
+            dap_time_to_str_rfc822( buf, sizeof(buf), current_time );
+
+            dap_http_header_add( &l_http_client->out_headers,"Date", buf );
 
             log_it( L_INFO," HTTP response with %u status code", l_http_client->reply_status_code );
-            dap_events_socket_write_f_unsafe(cl, "HTTP/1.1 %u %s\r\n",l_http_client->reply_status_code, l_http_client->reply_reason_phrase[0] ?
+            dap_events_socket_write_f_unsafe(a_esocket, "HTTP/1.1 %u %s\r\n",l_http_client->reply_status_code, l_http_client->reply_reason_phrase[0] ?
                             l_http_client->reply_reason_phrase : http_status_reason_phrase(l_http_client->reply_status_code) );
-            dap_events_socket_set_writable_unsafe(cl, true);
-            dap_http_client_out_header_generate( l_http_client );
+            dap_events_socket_set_writable_unsafe(a_esocket, true);
             l_http_client->state_write = DAP_HTTP_CLIENT_STATE_HEADERS;
         } break;
 
@@ -506,33 +532,48 @@ void dap_http_client_write( dap_events_socket_t * cl, void *arg )
             if ( hdr == NULL ) {
                 log_it(L_DEBUG, "Output: headers are over (reply status code %u content_lentgh %u)",
                        l_http_client->reply_status_code, l_http_client->out_content_length);
-                dap_events_socket_write_f_unsafe(cl, "\r\n");
-                dap_events_socket_set_writable_unsafe(cl, true);
+                dap_events_socket_write_f_unsafe(a_esocket, "\r\n");
+                dap_events_socket_set_writable_unsafe(a_esocket, true);
                 if ( l_http_client->out_content_length || l_http_client->out_content_ready ) {
                     l_http_client->state_write=DAP_HTTP_CLIENT_STATE_DATA;
                 } else {
                     log_it( L_DEBUG, "Nothing to output" );
                     l_http_client->state_write = DAP_HTTP_CLIENT_STATE_NONE;
-                    dap_events_socket_set_writable_unsafe( cl, false );
-                    cl->flags |= DAP_SOCK_SIGNAL_CLOSE;
+                    dap_events_socket_set_writable_unsafe( a_esocket, false );
+                    a_esocket->flags |= DAP_SOCK_SIGNAL_CLOSE;
                 }
-                dap_events_socket_set_readable_unsafe( cl, true );
+                dap_events_socket_set_readable_unsafe( a_esocket, true );
             } else {
                 //log_it(L_WARNING,"Output: header %s: %s",hdr->name,hdr->value);
-                dap_events_socket_write_f_unsafe(cl, "%s: %s\r\n", hdr->name, hdr->value);
-                dap_events_socket_set_writable_unsafe(cl, true);
+                dap_events_socket_write_f_unsafe(a_esocket, "%s: %s\r\n", hdr->name, hdr->value);
+                dap_events_socket_set_writable_unsafe(a_esocket, true);
                 dap_http_header_remove( &l_http_client->out_headers, hdr );
             }
         } break;
-        case DAP_HTTP_CLIENT_STATE_DATA:
-        {
-          if ( l_http_client->proc ){
-            if ( l_http_client->proc->data_write_callback ){
-                l_http_client->proc->data_write_callback( l_http_client, NULL );
+        case DAP_HTTP_CLIENT_STATE_DATA:{
+            if ( l_http_client->proc ){
+                pthread_rwlock_rdlock(&l_http_client->proc->cache_rwlock);
+                if ( l_http_client->proc->cache == NULL && l_http_client->proc->data_write_callback ){
+                    pthread_rwlock_unlock(&l_http_client->proc->cache_rwlock);
+                    l_http_client->proc->data_write_callback( l_http_client, NULL );
+                }else if(l_http_client->proc->cache) {
+                    size_t l_to_send=l_http_client->proc->cache->body_size-l_http_client->out_cache_position ;
+                    size_t l_sent = dap_events_socket_write_unsafe(l_http_client->esocket,
+                                                   l_http_client->proc->cache->body+l_http_client->out_cache_position,
+                                                   l_to_send );
+                    if(l_sent){
+                        if ( l_http_client->out_cache_position + l_sent >= l_http_client->proc->cache->body_size ){ // All is sent
+                            l_http_client->esocket->flags |= DAP_SOCK_SIGNAL_CLOSE;
+                            l_http_client->state_write = DAP_HTTP_CLIENT_STATE_NONE;
+                            dap_events_socket_set_writable_unsafe( a_esocket, false );
+                        }else
+                            l_http_client->out_cache_position += l_sent;
+                    }
+                    pthread_rwlock_unlock(&l_http_client->proc->cache_rwlock);
+                }
+            }else{
+                log_it(L_WARNING, "No http proc, nothing to write");
             }
-          }else{
-              log_it(L_WARNING, "No http proc, nothing to write");
-          }
         }
         break;
   }
@@ -545,10 +586,6 @@ void dap_http_client_write( dap_events_socket_t * cl, void *arg )
 void dap_http_client_out_header_generate(dap_http_client_t *cl_ht)
 {
   char buf[1024];
-  time_t current_time = time( NULL );
-  dap_time_to_str_rfc822( buf, sizeof(buf), current_time );
-
-  dap_http_header_add( &cl_ht->out_headers,"Date", buf );
 
   if ( cl_ht->reply_status_code == 200 ) {
 
diff --git a/dap-sdk/net/server/http_server/http_client/dap_http_header.c b/dap-sdk/net/server/http_server/http_client/dap_http_header.c
index e2458cadef..59ee805194 100644
--- a/dap-sdk/net/server/http_server/http_client/dap_http_header.c
+++ b/dap-sdk/net/server/http_server/http_client/dap_http_header.c
@@ -1,24 +1,26 @@
 /*
- Copyright (c) 2017-2018 (c) Project "DeM Labs Inc" https://github.com/demlabsinc
-  All rights reserved.
+ * Authors:
+ * Dmitriy A. Gearasimov <gerasimov.dmitriy@demlabs.net>
+ * DeM Labs Ltd.   https://demlabs.net
+ * Copyright  (c) 2021
+ * All rights reserved.
 
- This file is part of DAP (Deus Applications Prototypes) the open source project
+ This file is part of DAP SDK the open source project
 
-    DAP (Deus Applicaions Prototypes) is free software: you can redistribute it and/or modify
-    it under the terms of the GNU Lesser General Public License as published by
+    DAP SDK is free software: you can redistribute it and/or modify
+    it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.
 
-    DAP is distributed in the hope that it will be useful,
+    DAP SDK is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU Lesser General Public License for more details.
+    GNU General Public License for more details.
 
-    You should have received a copy of the GNU Lesser General Public License
-    along with any DAP based project.  If not, see <http://www.gnu.org/licenses/>.
+    You should have received a copy of the GNU General Public License
+    along with any DAP SDK based project.  If not, see <http://www.gnu.org/licenses/>.
 */
 
-
 #include <stdio.h>
 #include <stdarg.h>
 #include <string.h>
@@ -32,8 +34,10 @@
 #endif
 
 #include <pthread.h>
+#include <utlist.h>
 
 #include "dap_common.h"
+#include "dap_strfuncs.h"
 #include "dap_events_socket.h"
 #include "dap_http_client.h"
 #include "dap_http_header.h"
@@ -215,3 +219,20 @@ dap_http_header_t *dap_http_header_find( dap_http_header_t *top, const char *nam
 
   return ret;
 }
+
+/**
+ * @brief dap_http_headers_dup
+ * @param a_top
+ * @return
+ */
+dap_http_header_t * dap_http_headers_dup(dap_http_header_t * a_top)
+{
+    dap_http_header_t * l_hdr=NULL, * l_ret = NULL;
+    DL_FOREACH(a_top,l_hdr){
+        dap_http_header_t * l_hdr_copy = DAP_NEW_Z(dap_http_header_t);
+        l_hdr_copy->name = dap_strdup(l_hdr->name);
+        l_hdr_copy->value = dap_strdup(l_hdr->value);
+        DL_APPEND(l_ret,l_hdr_copy);
+    }
+    return l_ret;
+}
diff --git a/dap-sdk/net/server/http_server/http_client/include/dap_http_client.h b/dap-sdk/net/server/http_server/http_client/include/dap_http_client.h
index 864233216d..2be93454e1 100644
--- a/dap-sdk/net/server/http_server/http_client/include/dap_http_client.h
+++ b/dap-sdk/net/server/http_server/http_client/include/dap_http_client.h
@@ -62,6 +62,7 @@ typedef struct dap_http_client
     char out_content_type[256];
     time_t out_last_modified;
     bool out_connection_close;
+    size_t out_cache_position;
 
     dap_events_socket_t *esocket;
     struct dap_http * http;
@@ -82,15 +83,15 @@ typedef struct dap_http_client
 extern "C" {
 #endif
 
-int dap_http_client_init( );
-void dap_http_client_deinit( );
-void dap_http_client_new( dap_events_socket_t * cl,void *arg ); // Creates HTTP client's internal structure
-void dap_http_client_delete( dap_events_socket_t * cl,void *arg ); // Free memory for HTTP client's internal structure
+int dap_http_client_init(void);
+void dap_http_client_deinit(void);
+void dap_http_client_new( dap_events_socket_t *a_esocket, void *a_arg ); // Creates HTTP client's internal structure
+void dap_http_client_delete( dap_events_socket_t * a_esocket,void *a_arg ); // Free memory for HTTP client's internal structure
 
-void dap_http_client_read( dap_events_socket_t * cl,void *arg ); // Process read event
-void dap_http_client_write( dap_events_socket_t * cl,void *arg ); // Process write event
-void dap_http_client_error( dap_events_socket_t * cl,int arg ); // Process error event
-void dap_http_client_out_header_generate( dap_http_client_t *cl_ht );
+void dap_http_client_read( dap_events_socket_t * a_esocket,void *a_arg ); // Process read event
+void dap_http_client_write( dap_events_socket_t * a_esocket,void *a_arg ); // Process write event
+void dap_http_client_error( dap_events_socket_t * a_esocket,int a_arg ); // Process error event
+void dap_http_client_out_header_generate( dap_http_client_t *a_http_client );
 
 #ifdef __cplusplus
 }
diff --git a/dap-sdk/net/server/http_server/http_client/include/dap_http_header.h b/dap-sdk/net/server/http_server/http_client/include/dap_http_header.h
index 55c3e37054..f3b1e8e284 100644
--- a/dap-sdk/net/server/http_server/http_client/include/dap_http_header.h
+++ b/dap-sdk/net/server/http_server/http_client/include/dap_http_header.h
@@ -1,27 +1,27 @@
 /*
- Copyright (c) 2017-2018 (c) Project "DeM Labs Inc" https://github.com/demlabsinc
-  All rights reserved.
+ * Authors:
+ * Dmitriy A. Gearasimov <gerasimov.dmitriy@demlabs.net>
+ * DeM Labs Ltd.   https://demlabs.net
+ * Copyright  (c) 2021
+ * All rights reserved.
 
- This file is part of DAP (Deus Applications Prototypes) the open source project
+ This file is part of DAP SDK the open source project
 
-    DAP (Deus Applicaions Prototypes) is free software: you can redistribute it and/or modify
-    it under the terms of the GNU Lesser General Public License as published by
+    DAP SDK is free software: you can redistribute it and/or modify
+    it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.
 
-    DAP is distributed in the hope that it will be useful,
+    DAP SDK is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU Lesser General Public License for more details.
+    GNU General Public License for more details.
 
-    You should have received a copy of the GNU Lesser General Public License
-    along with any DAP based project.  If not, see <http://www.gnu.org/licenses/>.
+    You should have received a copy of the GNU General Public License
+    along with any DAP SDK based project.  If not, see <http://www.gnu.org/licenses/>.
 */
 
-
-#ifndef _DAP_HTTP_HEADER_H_
-#define _DAP_HTTP_HEADER_H_
-
+#pragma once
 //Structure for holding HTTP header in the bidirectional list
 typedef struct dap_http_header{
     char *name;
@@ -32,21 +32,22 @@ typedef struct dap_http_header{
 
 struct dap_http_client;
 
-extern int dap_http_header_init(); // Init module
-extern void dap_http_header_deinit(); // Deinit module
+int dap_http_header_init(); // Init module
+void dap_http_header_deinit(); // Deinit module
+
+int dap_http_header_parse(struct dap_http_client * cl_ht, const char * str);
 
-extern int dap_http_header_parse(struct dap_http_client * cl_ht, const char * str);
+dap_http_header_t * dap_http_header_add(dap_http_header_t ** top, const char*name, const char * value);
 
-extern dap_http_header_t * dap_http_header_add(dap_http_header_t ** top, const char*name, const char * value);
+dap_http_header_t * dap_http_out_header_add(struct dap_http_client * ht, const char*name, const char * value);
+dap_http_header_t * dap_http_out_header_add_f(struct dap_http_client * ht, const char*name, const char * value,...);
 
-extern dap_http_header_t * dap_http_out_header_add(struct dap_http_client * ht, const char*name, const char * value);
-extern dap_http_header_t * dap_http_out_header_add_f(struct dap_http_client * ht, const char*name, const char * value,...);
+dap_http_header_t * dap_http_header_find(dap_http_header_t * top, const char*name);
 
-extern dap_http_header_t * dap_http_header_find(dap_http_header_t * top, const char*name);
+dap_http_header_t * dap_http_headers_dup(dap_http_header_t * a_top);
 
-extern void dap_http_header_remove(dap_http_header_t ** top,dap_http_header_t * hdr );
+void dap_http_header_remove(dap_http_header_t ** top,dap_http_header_t * hdr );
 
 // For debug output
-extern void print_dap_http_headers(dap_http_header_t * top);
+void print_dap_http_headers(dap_http_header_t * top);
 
-#endif
diff --git a/dap-sdk/net/server/http_server/include/dap_http.h b/dap-sdk/net/server/http_server/include/dap_http.h
index 0d1fcfb84f..257b73e009 100644
--- a/dap-sdk/net/server/http_server/include/dap_http.h
+++ b/dap-sdk/net/server/http_server/include/dap_http.h
@@ -26,6 +26,7 @@ See more details here <http://www.gnu.org/licenses/>.
 #include "dap_events_socket.h"
 #include "dap_http_header.h"
 #include "dap_http_client.h"
+#include "dap_http_cache.h"
 #include "uthash.h"
 
 struct dap_http;
@@ -37,6 +38,9 @@ typedef struct dap_http_url_proc{
     char url[512]; // First part of URL that will be processed
     struct dap_http * http; // Pointer to HTTP server instance
 
+    dap_http_cache_t * cache; // In memory cache, could be present or not
+    pthread_rwlock_t cache_rwlock;
+
     dap_http_client_callback_t new_callback; // Init internal structure
     dap_http_client_callback_t delete_callback; // Delete internal structure
 
@@ -77,3 +81,4 @@ void dap_http_add_proc(dap_http_t *sh, const char *url_path, void *internal
                              ,dap_http_client_callback_t data_write_callback
                              ,dap_http_client_callback_error_t error_callback ); // Add custom procesor for the HTTP server
 
+void dap_http_url_proc_cache_reset(dap_http_url_proc_t *a_url_proc);
diff --git a/dap-sdk/net/server/http_server/include/dap_http_cache.h b/dap-sdk/net/server/http_server/include/dap_http_cache.h
new file mode 100644
index 0000000000..c1e55a8d0a
--- /dev/null
+++ b/dap-sdk/net/server/http_server/include/dap_http_cache.h
@@ -0,0 +1,40 @@
+/*
+ * Authors:
+ * Dmitriy A. Gearasimov <gerasimov.dmitriy@demlabs.net>
+ * DeM Labs Ltd.   https://demlabs.net
+ * Copyright  (c) 2021
+ * All rights reserved.
+
+ This file is part of DAP SDK the open source project
+
+    DAP SDK is free software: you can redistribute it and/or modify
+    it under the terms of the GNU General Public License as published by
+    the Free Software Foundation, either version 3 of the License, or
+    (at your option) any later version.
+
+    DAP SDK is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+    GNU General Public License for more details.
+
+    You should have received a copy of the GNU General Public License
+    along with any DAP SDK based project.  If not, see <http://www.gnu.org/licenses/>.
+*/
+#pragma once
+#include "dap_common.h"
+#include "dap_http_header.h"
+
+// Cache object
+typedef struct dap_http_cache
+{
+    struct dap_http_url_proc * url_proc;
+    byte_t *body;
+    size_t body_size;
+    dap_http_header_t * headers;
+
+    time_t ts_expire;
+} dap_http_cache_t;
+
+dap_http_cache_t * dap_http_cache_update(struct dap_http_url_proc * a_url_proc, byte_t * a_body, size_t a_body_size,
+                                         dap_http_header_t * a_headers, time_t ts_expire );
+void dap_http_cache_delete(dap_http_cache_t * a_http_cache);
diff --git a/dap-sdk/net/server/http_server/include/dap_http_simple.h b/dap-sdk/net/server/http_server/include/dap_http_simple.h
index bd2ccc1a51..a0d1534035 100644
--- a/dap-sdk/net/server/http_server/include/dap_http_simple.h
+++ b/dap-sdk/net/server/http_server/include/dap_http_simple.h
@@ -65,6 +65,7 @@ typedef struct dap_http_simple {
 #define DAP_HTTP_SIMPLE(a) ((dap_http_simple_t*) (a)->_inheritor )
 
 void dap_http_simple_proc_add( dap_http_t *sh, const char *url_path, size_t reply_size_max, dap_http_simple_callback_t cb ); // Add simple processor
+
 int  dap_http_simple_module_init( void );
 void dap_http_simple_module_deinit(void);
 
@@ -83,4 +84,5 @@ void dap_http_simple_set_pass_unknown_user_agents( bool pass );
 
 size_t dap_http_simple_reply( dap_http_simple_t *shs, void *data, size_t data_size );
 size_t dap_http_simple_reply_f( dap_http_simple_t *shs, const char *data, ... );
+dap_http_cache_t * dap_http_simple_make_cache_from_reply(dap_http_simple_t * a_http_simple , time_t a_ts_expire );
 
diff --git a/dap-sdk/net/server/http_server/dap_http.c b/dap-sdk/net/server/http_server/dap_http.c
index d28c653537..04458bfb29 100644
--- a/dap-sdk/net/server/http_server/dap_http.c
+++ b/dap-sdk/net/server/http_server/dap_http.c
@@ -52,7 +52,7 @@
 #include "dap_http_header.h"
 #include "dap_http_client.h"
 
-#define LOG_TAG "dap_http"
+#define LOG_TAG "http"
 
 
 /**
@@ -61,18 +61,18 @@
  */
 int dap_http_init( )
 {
-  if ( dap_http_header_init() != 0 ) { // Init submodule for headers manipulations
-    log_it(L_CRITICAL,"Can't init HTTP headers processing submodule");
-    return -1;
-  }
-
-  if ( dap_http_client_init() !=0 ) { // Init submodule for HTTP client event processing
-    log_it(L_CRITICAL,"Can't init HTTP client submodule");
-    return -2;
-  }
-
-  log_it( L_NOTICE, "Initialized HTTP server module" );
-  return 0;
+    if ( dap_http_header_init() != 0 ) { // Init submodule for headers manipulations
+        log_it(L_CRITICAL,"Can't init HTTP headers processing submodule");
+        return -1;
+    }
+
+    if ( dap_http_client_init() !=0 ) { // Init submodule for HTTP client event processing
+        log_it(L_CRITICAL,"Can't init HTTP client submodule");
+        return -2;
+    }
+
+    log_it( L_NOTICE, "Initialized HTTP server module" );
+    return 0;
 }
 
 /**
@@ -80,32 +80,33 @@ int dap_http_init( )
  */
 void dap_http_deinit()
 {
-  dap_http_header_deinit( );
-  dap_http_client_deinit( );
+    dap_http_header_deinit( );
+    dap_http_client_deinit( );
 }
 
 
 /**
  * @brief dap_server_http_init Init HTTP server
  * @param sh Server instance
  * @return 0 if ok lesser number if error
  */
-int dap_http_new( dap_server_t *sh, const char * server_name )
+int dap_http_new( dap_server_t *a_server, const char * a_server_name )
 {
-  sh->_inheritor = calloc( 1, sizeof(dap_http_t) );
+    a_server->_inheritor = DAP_NEW_Z(dap_http_t);
 
-  dap_http_t *shttp = DAP_HTTP( sh );
+    dap_http_t *l_http = DAP_HTTP( a_server );
 
-  shttp->server = sh;
-  strncpy( shttp->server_name, server_name, sizeof(shttp->server_name)-1 );
+    l_http->server = a_server;
+    strncpy( l_http->server_name, a_server_name, sizeof(l_http->server_name)-1 );
 
-  sh->client_callbacks.new_callback    = dap_http_client_new;
-  sh->client_callbacks.delete_callback = dap_http_client_delete;
-  sh->client_callbacks.read_callback   = dap_http_client_read;
-  sh->client_callbacks.write_callback  = dap_http_client_write;
-  sh->client_callbacks.error_callback  = dap_http_client_error;
+    a_server->client_callbacks.new_callback    = dap_http_client_new;
+    a_server->client_callbacks.delete_callback = dap_http_client_delete;
+    a_server->client_callbacks.read_callback   = dap_http_client_read;
+    a_server->client_callbacks.write_callback  = dap_http_client_write;
+    a_server->client_callbacks.error_callback  = dap_http_client_error;
 
-  return 0;
+    return 0;
 }
 
 /**
@@ -113,59 +114,57 @@ int dap_http_new( dap_server_t *sh, const char * server_name )
  * @param sh Server's instance
  * @param arg Non-used argument
  */
-void dap_http_delete( dap_server_t *sh, void * arg )
+void dap_http_delete( dap_server_t *a_server, void * a_arg )
 {
-  (void) arg;
-  (void) sh;
-  dap_http_t *shttp = DAP_HTTP( sh );
-  dap_http_url_proc_t *up, *tmp;
-
-  HASH_ITER( hh, shttp->url_proc ,up, tmp ) {
-    HASH_DEL(shttp->url_proc, up);
-    if( up->_inheritor )
-      free( up->_inheritor );
-    free( up );
-  }
+    (void) a_arg;
+    dap_http_t *l_http = DAP_HTTP( a_server );
+    dap_http_url_proc_t *l_url_proc, *l_tmp;
+
+    HASH_ITER( hh, l_http->url_proc ,l_url_proc, l_tmp ) {
+        HASH_DEL(l_http->url_proc, l_url_proc);
+        if( l_url_proc->_inheritor )
+            DAP_DELETE(l_url_proc->_inheritor );
+        DAP_DELETE(l_url_proc );
+    }
 }
 
 
 /**
  * @brief dap_http_add_proc  Add custom procesor for the HTTP server
  * @param sh                Server's instance
  * @param url_path          Part of URL to be processed
  * @param read_callback     Callback for read in DATA state
  * @param write_callback    Callback for write in DATA state
  * @param error_callback    Callback for error processing
  */
-void dap_http_add_proc(dap_http_t *sh, const char *url_path, void *internal
-                      ,dap_http_client_callback_t new_callback
-                      ,dap_http_client_callback_t delete_callback
-                      ,dap_http_client_callback_t headers_read_callback
-                      ,dap_http_client_callback_t headers_write_callback
-                      ,dap_http_client_callback_t data_read_callback
-                      ,dap_http_client_callback_t data_write_callback
-                      ,dap_http_client_callback_error_t error_callback
+void dap_http_add_proc(dap_http_t *a_http, const char *a_url_path, void *a_inheritor
+                      ,dap_http_client_callback_t a_new_callback
+                      ,dap_http_client_callback_t a_delete_callback
+                      ,dap_http_client_callback_t a_headers_read_callback
+                      ,dap_http_client_callback_t a_headers_write_callback
+                      ,dap_http_client_callback_t a_data_read_callback
+                      ,dap_http_client_callback_t a_data_write_callback
+                      ,dap_http_client_callback_error_t a_error_callback
 
                       )
 {
-  dap_http_url_proc_t *up = (dap_http_url_proc_t *) calloc( 1, sizeof(dap_http_url_proc_t) );
+    dap_http_url_proc_t *l_url_proc = DAP_NEW_Z(dap_http_url_proc_t);
 
-  strncpy( up->url, url_path, sizeof(up->url)-1 );
+    strncpy( l_url_proc->url, a_url_path, sizeof(l_url_proc->url)-1 );
 
-  up->new_callback    = new_callback;
-  up->delete_callback = delete_callback;
+    l_url_proc->new_callback    = a_new_callback;
+    l_url_proc->delete_callback = a_delete_callback;
 
-  up->data_read_callback = data_read_callback;
-  up->data_write_callback = data_write_callback;
-  up->headers_read_callback = headers_read_callback;
-  up->headers_write_callback = headers_write_callback;
-  up->error_callback = error_callback;
+    l_url_proc->data_read_callback = a_data_read_callback;
+    l_url_proc->data_write_callback = a_data_write_callback;
+    l_url_proc->headers_read_callback = a_headers_read_callback;
+    l_url_proc->headers_write_callback = a_headers_write_callback;
+    l_url_proc->error_callback = a_error_callback;
 
-  up->_inheritor = internal;
+    l_url_proc->_inheritor = a_inheritor;
 
-  HASH_ADD_STR( sh->url_proc, url, up );
+    HASH_ADD_STR( a_http->url_proc, url, l_url_proc );
 
-  log_it( L_DEBUG, "Added URL processor for '%s' path", up->url );
+    log_it( L_DEBUG, "Added URL processor for '%s' path", l_url_proc->url );
 }
 
-
@@ -80,32 +80,33 @@ int dap_http_init( )
 
 /**
- * @brief dap_server_http_init Init HTTP server
- * @param sh Server instance
+ * @brief dap_server_http_init   Init HTTP server
+ * @param a_server               Server instance
+ * @param a_server_name          Server name
  * @return 0 if ok lesser number if error
  */
@@ -113,59 +114,57 @@ int dap_http_new( dap_server_t *sh, const char * server_name )
 
 /**
- * @brief dap_http_add_proc  Add custom procesor for the HTTP server
- * @param sh                Server's instance
- * @param url_path          Part of URL to be processed
- * @param read_callback     Callback for read in DATA state
- * @param write_callback    Callback for write in DATA state
- * @param error_callback    Callback for error processing
+ * @brief dap_http_add_proc   Add custom procesor for the HTTP server
+ * @param a_http              Server's instance
+ * @param a_url_path          Part of URL to be processed
+ * @param a_read_callback     Callback for read in DATA state
+ * @param a_write_callback    Callback for write in DATA state
+ * @param a_error_callback    Callback for error processing
  */
-- 
GitLab